/*  $Id: xml_tokenizer.cpp,v 1.15 2003/09/18 18:03:52 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Core/precomp.h"
#include "API/Core/XML/xml_tokenizer.h"
#include "API/Core/XML/xml_token.h"
#include "API/Core/System/clanstring.h"
#include "API/Core/System/error.h"
#include "xml_tokenizer_generic.h"

std::string replace_escapes(std::string str);

/////////////////////////////////////////////////////////////////////////////
// CL_XMLTokenizer construction:

CL_XMLTokenizer::CL_XMLTokenizer() : impl(0)
{
}

CL_XMLTokenizer::CL_XMLTokenizer(const CL_XMLTokenizer &copy) : impl(copy.impl)
{
	if (impl) impl->add_ref();
}

CL_XMLTokenizer::CL_XMLTokenizer(CL_InputSource *input, bool delete_input) : impl(new CL_XMLTokenizer_Generic)
{
	impl->input = input;
	impl->delete_input = delete_input;
	impl->size = input->size();
	impl->data.resize(impl->size);
	input->read(&impl->data[0], impl->size);
	impl->pos = 0;
	impl->add_ref();
}

CL_XMLTokenizer::~CL_XMLTokenizer()
{
	if (impl) impl->release_ref();
}

/////////////////////////////////////////////////////////////////////////////
// CL_XMLTokenizer attributes:

/////////////////////////////////////////////////////////////////////////////
// CL_XMLTokenizer operations:

CL_XMLTokenizer &CL_XMLTokenizer::operator =(const CL_XMLTokenizer &copy)
{
	if (impl) impl->release_ref();
	impl = copy.impl;
	if (impl) impl->add_ref();
	return *this;
}

CL_XMLToken CL_XMLTokenizer::next()
{
	if (impl == 0) return CL_XMLToken();

	if (impl->pos == impl->size) return CL_XMLToken(); // EOF, return null token.

	if (impl->data[impl->pos] != '<') // Text node
	{
		std::string::size_type start_pos = impl->pos;
		std::string::size_type end_pos = impl->data.find('<', start_pos);
		if (end_pos == impl->data.npos) end_pos = impl->size;
		impl->pos = end_pos;

		CL_XMLToken token;
		token.set_type(CL_XMLToken::TEXT_TOKEN);
		token.set_value(replace_escapes(impl->data.substr(start_pos, end_pos-start_pos)));
		return token;
	}
	else // Tag node
	{
		impl->pos++;
		if (impl->pos == impl->size) throw CL_Error("Premature end of XML data!");

		// Try to early predict what sort of node it might be:
		bool closing = false;
		bool questionMark = false;
		bool exclamationMark = false;
		if (impl->data[impl->pos] == '/') closing = true;
		else if (impl->data[impl->pos] == '?') questionMark = true;
		else if (impl->data[impl->pos] == '!') exclamationMark = true;
		if (closing || questionMark || exclamationMark)
		{
			impl->pos++;
			if (impl->pos == impl->size) throw CL_Error("Premature end of XML data!");
		}

		if (exclamationMark) // check for cdata section or comments
		{
			if (impl->data.compare(impl->pos, 2, "--") == 0) // comment block
			{
				std::string::size_type start_pos = impl->pos+2;
				std::string::size_type end_pos = impl->data.find("-->", start_pos);
				if (end_pos == impl->data.npos) throw CL_Error("Premature end of XML data!");
				impl->pos = end_pos+3;

				CL_XMLToken token;
				token.set_type(CL_XMLToken::COMMENT_TOKEN);
				token.set_variant(CL_XMLToken::SINGLE);
				token.set_value(replace_escapes(impl->data.substr(start_pos, end_pos-start_pos)));
				return token;
			}

			if (impl->data.compare(impl->pos, 7, "[CDATA[") != 0) throw CL_Error(CL_String::format("Error in XML stream at position %1", static_cast<int>(impl->pos)));
			std::string::size_type start_pos = impl->pos+7;
			std::string::size_type end_pos = impl->data.find("]]>", start_pos);
			if (end_pos == impl->data.npos) throw CL_Error("Premature end of XML data!");
			impl->pos = end_pos+3;

			CL_XMLToken token;
			token.set_type(CL_XMLToken::CDATA_SECTION_TOKEN);
			token.set_variant(CL_XMLToken::SINGLE);
			token.set_value(replace_escapes(impl->data.substr(start_pos, end_pos-start_pos)));
			return token;
		}

		// Extract the tag name:
		std::string::size_type start_pos = impl->pos;
		std::string::size_type end_pos = impl->data.find_first_of(" \r\n\t?/>", start_pos);
		if (end_pos == impl->data.npos) throw CL_Error("Premature end of XML data!");
		impl->pos = end_pos;

		CL_XMLToken token;
		token.set_type(questionMark ? CL_XMLToken::PROCESSING_INSTRUCTION_TOKEN : CL_XMLToken::ELEMENT_TOKEN);
		token.set_variant(closing ? CL_XMLToken::END : CL_XMLToken::BEGIN);
		token.set_name(replace_escapes(impl->data.substr(start_pos, end_pos-start_pos)));

		// Check for possible attributes:
		while (true)
		{
			// Strip whitespace:
			impl->pos = impl->data.find_first_not_of(" \r\n\t", impl->pos);
			if (impl->pos == impl->data.npos) throw CL_Error("Premature end of XML data!");

			// End of tag, stop searching for more attributes:
			if (impl->data[impl->pos] == '/' || impl->data[impl->pos] == '?' || impl->data[impl->pos] == '>') break;

			// Extract attribute name:
			std::string::size_type start_pos = impl->pos;
			std::string::size_type end_pos = impl->data.find_first_of(" \r\n\t=", start_pos);
			if (end_pos == impl->data.npos) throw CL_Error("Premature end of XML data!");
			impl->pos = end_pos;

			std::string attributeName = impl->data.substr(start_pos, end_pos-start_pos);

			// Find seperator:
			impl->pos = impl->data.find_first_not_of(" \r\n\t", impl->pos);
			if (impl->pos == impl->data.npos || impl->pos == impl->size-1) throw CL_Error("Premature end of XML data!");
			if (impl->data[impl->pos++] != '=') throw CL_Error(CL_String::format("XML error(s), parser confused at line %1 (tag=%2, attributeName=%3)", impl->get_line_number(), token.get_name(), attributeName));

			// Strip whitespace:
			impl->pos = impl->data.find_first_not_of(" \r\n\t", impl->pos);
			if (impl->pos == impl->data.npos) throw CL_Error("Premature end of XML data!");

			// Extract attribute value:
			std::string first_of = " \r\n\t";
			if (impl->data[impl->pos] == '"') first_of = "\"";
			else if (impl->data[impl->pos] == '\'') first_of = "'";
			if (first_of != " \r\n\t")
			{
				impl->pos++;
				if (impl->pos == impl->size) throw CL_Error("Premature end of XML data!");
			}

			start_pos = impl->pos;
			end_pos = impl->data.find_first_of(first_of, start_pos);
			if (end_pos == impl->data.npos) throw CL_Error("Premature end of XML data!");
			std::string attributeValue = impl->data.substr(start_pos, end_pos-start_pos);

			impl->pos = end_pos + 1;
			if (impl->pos == impl->size) throw CL_Error("Premature end of XML data!");

			// Finally apply attribute to token:
			token.set_attribute(replace_escapes(attributeName), replace_escapes(attributeValue));
		}

		// Check if its singular:
		if (impl->data[impl->pos] == '/' || impl->data[impl->pos] == '?')
		{
			token.set_variant(CL_XMLToken::SINGLE);
			impl->pos++;
			if (impl->pos == impl->size) throw CL_Error("Premature end of XML data!");
		}

		// Data stream should be ending now.
		if (impl->data[impl->pos] != '>') throw CL_Error(CL_String::format("Error in XML stream, line %1 (expected end of tag)", impl->get_line_number()));
		impl->pos++;

		return token;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CL_XMLTokenizer implementation:

std::string replace_escapes(std::string str)
{
	std::string::size_type pos;

	pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find("&amp;", pos);
		if (pos == std::string::npos) break;
		str.replace( pos, 5, "&" );
		pos++;
	}
	
	pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find("&quot;", pos);
		if (pos == std::string::npos) break;
		str.replace( pos, 6, "\"" );
		pos++;
	}

	pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find("&apos;", pos);
		if (pos == std::string::npos) break;
		str.replace( pos, 6, "\'" );
		pos++;
	}

	pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find("&lt;", pos);
		if (pos == std::string::npos) break;
		str.replace( pos, 4, "<" );
		pos++;
	}

	pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find("&gt;", pos);
		if (pos == std::string::npos) break;
		str.replace( pos, 4, ">" );
		pos++;
	}

	return str;
}
