/*  $Id: font_generic.h,v 1.25 2003/10/14 15:52:12 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef header_font_generic
#define header_font_generic

#if _MSC_VER > 1000
#pragma once
#endif

#include "API/Display/sprite.h"
#include "API/Core/Math/origin.h"
#include "API/Display/font.h"
#include "API/Core/Math/point.h"
#include "API/Core/System/clonable.h"
#include <string>
#include <map>
#include <utility> //For std::pair

class CL_GraphicContext;
class CL_GlyphBuffer;
class CL_Size;
class CL_Rect;

class CL_Font_Generic : public CL_Clonable
{
//! Construction:
public:
	CL_Font_Generic(); 
	CL_Font_Generic(const CL_Font_Generic &copy);
	
	virtual ~CL_Font_Generic();

//! Attributes:
public:
	//: Alters the percieved width of each glyph
	int width_offset;
	
	//: Alters the percieved height of each line
	int height_offset;
	
	//: Alignment
	CL_Origin trans_origin;
	int trans_x;
	int trans_y;
	
	//: A list of characters (other than the newline character) that seperate words
	std::string delims;
	
	//: If 0, variable width font. If another number, all characters are this width
	int fixed_width;
	
	//: Gets the width of a single glyph directly from the map, including scaling and offsets
	int get_width(unsigned char chr) const;
	
	//: Gets the height of the font, including scaling and offsets
	int get_height() const;
	
	//: Returns true if the given character has any associated glyph
	bool is_glyph(unsigned char chr) const;
	
	void get_scale(float& x, float& y) const {glyphs.get_scale(x, y);}
	float get_alpha() const {return glyphs.get_alpha();}
	void get_color(float& r, float& g, float& b, float& a) const {glyphs.get_color(r, g, b, a);}
	void get_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const {glyphs.get_blend_func(src, dest);}
	void get_rot_hotspot(CL_Origin &origin, int &x, int &y) const {glyphs.get_rotation_hotspot(origin, x, y);}
	
//! Operations:
public:
	//: Copy assignment.
	CL_Font_Generic &operator =(const CL_Font_Generic &copy);
	
	//: Draws a glyph at a given location.
	//param x, y: The upper-left coordinates of where to draw the glyph.
	//param chr: The character to draw the glyph of.
	//param ang: The amount to rotate the glyph by. The hotspot is set with CL_Font::set_glyph_rot_hotspot().
	//- <p> If there is no such glyph for that character, nothing is drawn. </p>
	void draw_glyph(int x, int y, unsigned char chr, float ang) const;
	
	//: Sets the ctg map to a given letters string, creates if it neccessary, increments ref count.
	void set_letters(const std::string &letters);
	
	//: Loads a system font. Should be called exactly once, immediately after default construction.
	virtual void create_sysfont(
		const std::string &font_name,
		const std::string &letter_chars,
		int height,
		int width,
		bool bold,
		bool italic,
		bool underline,
		bool strikeout) = 0;
	
	//: Loads a regular font. Should be called exactly once, immediately after default construction.
	void create_font(
		const CL_Sprite &letters,
		const std::string &letter_chars,
		int space_width = -1,
		bool monospace = false);
	
	void set_scale(float x, float y) {glyphs.set_scale(x, y);}
	void set_alpha(float a) {glyphs.set_alpha(a);}
	void set_color(float r, float g, float b, float a) {glyphs.set_color(r, g, b, a);}
	void set_blend_func(CL_BlendFunc src, CL_BlendFunc dest) {glyphs.set_blend_func(src, dest);}
	void set_rot_hotspot(CL_Origin origin, int x, int y) {glyphs.set_rotation_hotspot(origin, x, y);}
	
//! Implementation
private:
	//: Glyphs of the actual characters.
	mutable CL_Sprite glyphs;
	
	//: Height of font (i.e. height of the tallest character, before scaling/offsets)
	int height;
	
	//: Distance to go at any unrecognized character, before scaling/offsets
	int spacelen;
	
	//: Type for a glyph map vector and associated reference count
	//: Several are contained by a single static instance of CTGVector_Cache
	struct CTGVector_Cache;
	struct CTGVector
	{
		friend struct CTGVector_Cache;
		std::vector<short> char_to_glyph_map;
		unsigned int ref_count;
		CTGVector() : ref_count(1)
		{
			//Fill the vector with -1's, symbolizing unimplemented characters
			char_to_glyph_map.reserve(256);
			for (short i = 0; i < 256; ++i)
				char_to_glyph_map.push_back(-1);
		}
	};
	
	//: Current ctg map we're using, which matches our current letters string
	std::map<std::string, CTGVector>::iterator cur_ctg;
	
	//: To map characters to frames of the glyphs sprite: 
	//- <p> Key is the 'letters' string given to create_font),
	//- Value is the map from characters to glyphs, and the reference count. 
	//- We make a seperate wrapper class so that we can initialize the
	//- map with an entry for "" the first time its used.</p>
	struct CTGVector_Cache
	{
		std::map<std::string, CTGVector> cache;
		CTGVector_Cache() {cache.insert(std::pair<std::string, CTGVector>("", CTGVector()));}
	};
	
	static CTGVector_Cache ctg_vector_cache;
	
	//: Lowers reference count on current ctg map in the cache
	void decrement();
};

#endif
