/*  $Id: graphic_context_generic.h,v 1.9 2003/10/14 15:52:12 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef header_graphic_context_generic
#define header_graphic_context_generic

#if _MSC_VER > 1000
#pragma once
#endif

#include "API/signals.h"
#include "API/Core/Math/point.h"
#include "API/Core/Math/rect.h"

class CL_Color;
class CL_Gradient;

//: Implementation interface for the graphic context.
class CL_GraphicContext_Generic
{
//! Construction:
public:
	CL_GraphicContext_Generic();

	virtual ~CL_GraphicContext_Generic();

//! Attributes:
public:
	//: Returns the current width of the context.
	virtual int get_width() const=0;

	//: Returns the current height of the context.
	virtual int get_height() const=0;

	//: Returns the current clipping rectangle used on the graphic context.
	virtual const CL_Rect &get_cliprect() const { return cliprects.front(); }

	//: Returns the current effective x-axis translation offset.
	virtual int get_translate_x() const { return translates.front().x; }

	//: Returns the current effective y-axis translation offset.
	virtual int get_translate_y() const { return translates.front().y; }
	
	//: Returns the current effective y-axis translation offset.
	virtual CL_Color get_pixel(int x, int y) const=0;

// Operations:
public:
	void add_ref();

	void release_ref();

	//: Begin 3d rendering mode.
	void begin_3d();

	//: End 3d rendering mode.
	void end_3d();

	//: Flushes current rendering batch.
	//- <p>With the OpenGL target, this causes the graphic context to end its current
	//- active glBegin()/glEnd() pair, making it possible to call other OpenGL calls.</p>
	virtual void flush() { return; }

	//: Begin window 2D rendering mode.
	//- <p> This effectively loads a Orto 2D projection matrix and makes sure
	//- the device are using sane texture parameter values.</p>
	virtual void setup_2d()=0;

	//: End windows 2D rendering mode.
	//- <p> This restores the rendering environment back to what it was before
	//- setup_2d() was called. </p>
	virtual void setup_3d()=0;

	//: Called on before every draw call. Meant to select current rendering context.
	virtual void select_gc()=0;

	//: Draw a pixel at (x, y) using the specified color.
	virtual void draw_pixel(int x, int y, const CL_Color &color)=0;

	//: Draw a line from (x1, y1) to (x2, y2) using the specified color.
	virtual void draw_line(int x1, int y1, int x2, int y2, const CL_Color &color)=0;

	//: Draw a rectangle using the specified color.
	virtual void draw_rect(const CL_Rect &rect, const CL_Color &color)=0;

	//: Draw a gradient rectangle using the specified gradient.
	virtual void fill_rect(const CL_Rect &rect, const CL_Gradient &gradient)=0;

	//: Draw a filled rectangle using the specified color.
	virtual void fill_rect(const CL_Rect &rect, const CL_Color &color)=0;

	//: Clears the whole context using the specified color.
	virtual void clear(const CL_Color &color)=0;

	//: Set the current clipping rectangle.
	virtual void set_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to stack; then set clipping rectangle to the
	//: union of the current rectangle and the passed rectangle.
	virtual void push_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to the stack.
	virtual void push_cliprect();

	//: Pop current clipping rectangle from the stack.
	virtual void pop_cliprect();

	//: Sets the translate offset.
	//- <p> The new offset will disregard any previous offset's, but will not
	//- empty the translation stack. The new translation offset will affect
	//- any subsequent display operations on the current displaycard, by
	//- translating the position of the display operation with the offset. </p>
	virtual void set_translate_offset(int x, int y);
	
	//: Push translation offset onto translation stack.
	//- <p> This offset will affect any subsequent display operations on the current
	//- displaycard, by translating the position of the display operation with the offset.
	//- The offset will be offset by any previous offsets pushed onto the stack,
	//- eg. it inherits the previous offset. </p>
	virtual void push_translate_offset(int delta_x, int delta_y);
	
	virtual void push_translate_offset();
	
	//: Pops the last pushed translation offset from the translation offset stack.
	//- <p> If the stack is empty, nothing will happen, and if the last
	//- translation offset is popped, the translation offset will be set to 0,0. </p>
	virtual void pop_translate_offset();

//! Implementation:
protected:
	std::list<CL_Rect> cliprects;

	std::list<CL_Point> translates;

private:
	int count_3d;

	int ref_count;
};

#endif
