/*  $Id: pixel_buffer.cpp,v 1.35 2003/09/19 10:33:02 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "API/Display/pixel_buffer.h"
#include "pixel_buffer_generic.h"
#include "pixel_buffer_memory.h"
#include <Hermes/H_Conv.h>
#include <Hermes/H_Pal.h>
#include <Hermes/Hermes.h>

/////////////////////////////////////////////////////////////////////////////
// CL_PixelBuffer construction:

CL_PixelBuffer::CL_PixelBuffer(CL_PixelBuffer_Generic *impl)
: impl(impl)
{
	if (impl) impl->add_ref();
}

CL_PixelBuffer::CL_PixelBuffer(int width, int height, int pitch, const CL_PixelFormat &format, void *data)
: impl(new CL_PixelBuffer_Memory(width, height, pitch, format, data))
{
	impl->add_ref();
}

CL_PixelBuffer::CL_PixelBuffer(int width, int height, int pitch, const CL_PixelFormat &format, const CL_Palette &palette, void *data)
: impl(new CL_PixelBuffer_Memory(width, height, pitch, format, palette, data))
{
	impl->add_ref();
}

CL_PixelBuffer::CL_PixelBuffer()
: impl(0)
{
}

CL_PixelBuffer::CL_PixelBuffer(const CL_PixelBuffer &copy)
: impl(copy.impl)
{
	if (impl) impl->add_ref();
}

CL_PixelBuffer::~CL_PixelBuffer()
{
	if (impl) impl->release_ref();
}

/////////////////////////////////////////////////////////////////////////////
// CL_PixelBuffer attributes:

const CL_PixelFormat &CL_PixelBuffer::get_format() const
{
	return impl->format;
}

const CL_Palette &CL_PixelBuffer::get_palette() const
{
	return impl->palette;
}

int CL_PixelBuffer::get_width() const
{
	return impl->width;
}

int CL_PixelBuffer::get_height() const
{
	return impl->height;
}

unsigned int CL_PixelBuffer::get_pitch() const
{
	return impl->pitch;
}

CL_GraphicContext *CL_PixelBuffer::get_gc()
{
	return 0;
}

const CL_GraphicContext *CL_PixelBuffer::get_gc() const
{
	return 0;
}

void *CL_PixelBuffer::get_data()
{
	return impl->get_data();
}

/////////////////////////////////////////////////////////////////////////////
// CL_PixelBuffer operations:

CL_PixelBuffer &CL_PixelBuffer::operator =(const CL_PixelBuffer &copy)
{
	if (impl) impl->release_ref();
	impl = copy.impl;
	if (impl) impl->add_ref();
	return *this;
}

void CL_PixelBuffer::lock()
{
	impl->lock_count++;
	if (impl->lock_count == 1) impl->perform_lock();
}

void CL_PixelBuffer::unlock()
{
	impl->lock_count--;
	if (impl->lock_count == 0) impl->perform_unlock();
}

void CL_PixelBuffer::convert(CL_PixelBuffer *target)
{
	target->lock();

	convert(
		target->get_data(),
		target->get_format(),
		target->get_pitch(),
		CL_Rect(
			0, 0,
			target->get_width(), target->get_height()));

	target->unlock();
}

void CL_PixelBuffer::convert(
	void *buffer,
	const CL_PixelFormat &format,
	int destPitch,
	const CL_Rect &destRect,
	const CL_Rect &srcRect)
{
	lock();

	bool nullSrc = false;
	if (srcRect.left == 0 && srcRect.top == 0 && srcRect.right == 0 && srcRect.bottom == 0)
	{
		nullSrc = true;
	}

	// Using new Hermes functionality for inverting the source alpha
	
	// Setup Hermes converter handles:
	HermesHandle handle = Hermes_ConverterInstance(HERMES_CONVERT_NORMAL);
	cl_assert(handle != 0);

	HermesHandle palette = Hermes_PaletteInstance();
	cl_assert(palette != 0);

	const unsigned int *palette_data = get_palette();
	Hermes_PaletteSet(palette, (int32 *) palette_data);

	HermesFormat *dest_format = Hermes_FormatNewEx(
		format.get_depth(),
		format.get_red_mask(),
		format.get_green_mask(),
		format.get_blue_mask(),
		format.get_alpha_mask(),
		(format.get_depth() == 8) ? 1 : 0, // bool indexed
		format.has_colorkey() ? 1 : 0,     // bool colorkey
		format.get_colorkey());
	cl_assert(dest_format != 0);

	HermesFormat *src_format = Hermes_FormatNewEx(
		get_format().get_depth(),
		get_format().get_red_mask(),
		get_format().get_green_mask(),
		get_format().get_blue_mask(),
		get_format().get_alpha_mask(),
		(get_format().get_depth() == 8) ? 1 : 0,    // bool indexed
		get_format().has_colorkey() ? 1 : 0,  // bool colorkey
		get_format().get_colorkey());
	cl_assert(src_format != 0);

	// Ask Hermes to convert:
	int result;

	result = Hermes_ConverterRequest(handle, src_format, dest_format);
	cl_assert(result != 0);

	result = Hermes_ConverterPalette(handle, palette, palette);
	cl_assert(result != 0);

	result = Hermes_ConverterCopy(
		handle,
		get_data(),
		srcRect.left,
		srcRect.top,
		nullSrc ? get_width() : srcRect.get_width(),
		nullSrc ? get_height() : srcRect.get_height(),
		get_pitch(),
		buffer,
		destRect.left,
		destRect.top,
		destRect.get_width(),
		destRect.get_height(),
		destPitch);
	cl_assert(result != 0);

	// Free handles:
	Hermes_ConverterReturn(handle);
	Hermes_PaletteReturn(palette);
	Hermes_FormatFree(src_format);
	Hermes_FormatFree(dest_format);

	// Apply alpha channel on dest if source image didnt have alpha:
	// (Hermes workaround because HERMES_CONVERT_INVERSE_ALPHA doesnt work properly)

	if (get_format().get_alpha_mask() == 0 || get_format().get_depth() == 8)
	{
		unsigned int alpha_mask = format.get_alpha_mask();
		switch (format.get_depth())
		{
		case 16:
			{
				for (int y = destRect.top; y < destRect.bottom; y++)
				{
					unsigned short *scanline = (unsigned short *)
						((unsigned char *) buffer + y * destPitch);

					for (int x = destRect.left; x < destRect.right; x++)
					{
						scanline[x] |= alpha_mask;
					}
				}
			}
			break;
		case 32:
			{
				for (int y = destRect.top; y < destRect.bottom; y++)
				{
					unsigned int *scanline = (unsigned int *)
						((unsigned char *) buffer + y * destPitch);

					for (int x = destRect.left; x < destRect.right; x++)
					{
						scanline[x] |= alpha_mask;
					}
				}
			}
			break;
		}
	}

	// Colorkey convert support:
	if (get_format().has_colorkey() && srcRect.get_size() == destRect.get_size() &&
		(format.has_colorkey() || format.get_alpha_mask()))
	{
		// Create colorkey buffer:
		char *ckbuffer = new char[srcRect.get_width()*srcRect.get_height()];
		unsigned char *src8 = (unsigned char *) get_data();
		unsigned short *src16 = 0;
		unsigned int *src32 = 0;
		unsigned int srckey = get_format().get_colorkey();
		unsigned int srcand = get_format().get_red_mask()|get_format().get_green_mask()|get_format().get_blue_mask()|get_format().get_alpha_mask();
		int src_pitch = get_pitch();
		int x, y;
		switch (get_format().get_depth())
		{
		case 8:
			for (y=srcRect.top; y<srcRect.bottom; y++)
			{
				for (x=srcRect.left; x<srcRect.right; x++)
				{
					*(ckbuffer++) = ((src8[x]&srcand) == srckey) ? 1 : 0;
				}
				src8 += src_pitch;
			}
			ckbuffer -= srcRect.get_width()*srcRect.get_height();
			break;

		case 16:
			for (y=srcRect.top; y<srcRect.bottom; y++)
			{
				src16 = (unsigned short *) src8;
				for (x=srcRect.left; x<srcRect.right; x++)
				{
					*(ckbuffer++) = ((src16[x]&srcand) == srckey) ? 1 : 0;
				}
				src8 += src_pitch;
			}
			ckbuffer -= srcRect.get_width()*srcRect.get_height();
			break;

		case 24:
			for (y=srcRect.top; y<srcRect.bottom; y++)
			{
				for (x=srcRect.left; x<srcRect.right; x++)
				{
					int pixel = (src8[x*3+0] << 16) + (src8[x*3+1] << 8) + src8[x*3+2];
					*(ckbuffer++) = ((pixel&srcand) == srckey) ? 1 : 0;
				}
				src8 += src_pitch;
			}
			ckbuffer -= srcRect.get_width()*srcRect.get_height();
			break;

		case 32:
			for (y=srcRect.top; y<srcRect.bottom; y++)
			{
				src32 = (unsigned int *) src8;
				for (x=srcRect.left; x<srcRect.right; x++)
				{
					*(ckbuffer++) = ((src32[x]&srcand) == srckey) ? 1 : 0;
				}
				src8 += src_pitch;
			}
			ckbuffer -= srcRect.get_width()*srcRect.get_height();
			break;
		}

		unsigned char *dest8 = (unsigned char *) buffer;
		unsigned short *dest16 = 0;
		unsigned int *dest32 = 0;
		unsigned int destkey = format.get_colorkey();
		if (format.has_colorkey() == false) destkey = 0; // format.get_alpha_mask();
		int dest_pitch = destPitch;

		// Copy source colorkey to dest colorkey:
		switch (format.get_depth())
		{
		case 8:
			for (y=destRect.top; y<destRect.bottom; y++)
			{
				for (x=destRect.left; x<destRect.right; x++)
				{
					if (*(ckbuffer++)) dest8[x] = destkey;
				}
				dest8 += dest_pitch;
			}
			ckbuffer -= destRect.get_width()*destRect.get_height();
			break;

		case 16:
			for (y=destRect.top; y<destRect.bottom; y++)
			{
				dest16 = (unsigned short *) dest8;
				for (x=destRect.left; x<destRect.right; x++)
				{
					if (*(ckbuffer++)) dest16[x] = destkey;
				}
				dest8 += dest_pitch;
			}
			ckbuffer -= destRect.get_width()*destRect.get_height();
			break;

		case 24:
			for (y=destRect.top; y<destRect.bottom; y++)
			{
				for (x=destRect.left; x<destRect.right; x++)
				{
					if (*(ckbuffer++))
					{
						dest8[x*3+0] = ((destkey >> 16) & 0xff);
						dest8[x*3+1] = ((destkey >> 8) & 0xff);
						dest8[x*3+2] = (destkey & 0xff);
					}
				}
				dest8 += dest_pitch;
			}
			ckbuffer -= destRect.get_width()*destRect.get_height();
			break;

		case 32:
			for (y=destRect.top; y<destRect.bottom; y++)
			{
				dest32 = (unsigned int *) dest8;
				for (x=destRect.left; x<destRect.right; x++)
				{
					if (*(ckbuffer++)) dest32[x] = destkey;
				}
				dest8 += dest_pitch;
			}
			ckbuffer -= destRect.get_width()*destRect.get_height();
			break;
		}

		delete[] ckbuffer;
	}

	unlock();
}

void CL_PixelBuffer::convert_line(void *buffer, const CL_PixelFormat &format, int y)
{
	convert(
		buffer,
		format,
		(format.get_depth()+7)/8*get_width(),
		CL_Rect(0, 0, get_width(), 1),
		CL_Rect(0, y, get_width(), 1));
}
/*
class BlitParams
{
public:
	char *src_data;
	char *dest_data;
	unsigned int width;
	unsigned int height;
	unsigned int src_pitch;
	unsigned int dest_pitch;
	int src_red_shift;
	int src_green_shift;
	int src_blue_shift;
	int src_alpha_shift;
	int dest_red_shift;
	int dest_green_shift;
	int dest_blue_shift;
	int dest_alpha_shift;
	unsigned int src_red_mask;
	unsigned int src_green_mask;
	unsigned int src_blue_mask;
	unsigned int src_alpha_mask;
	unsigned int dest_red_mask;
	unsigned int dest_green_mask;
	unsigned int dest_blue_mask;
	unsigned int dest_alpha_mask;
};

template<class SrcType, class DestType>
void template_blitter(BlitParams &params)
{
	for (unsigned int y = 0; y<params.height; y++)
	{
		SrcType *src_data = (SrcType *) (params.src_data + params.src_pitch * y);
		DestType *dest_data = (DestType *) (params.dest_data + params.dest_pitch * y);

		for (unsigned int x = 0; x < params.width; x++)
		{
			unsigned int src_pixel = src_data[x];

			DestType result_pixel = src_pixel;
			// todo: someone make a god damn working blitter here!
			dest_data[x] = result_pixel;
		}
	}
}

static int get_mask_shift(unsigned int mask)
{
	int count = 0;
	while (mask > 0) { count++; mask >>= 1; }
	return count;
}

void CL_PixelBuffer::blit(CL_PixelBuffer *target, const CL_Rect &src, const CL_Rect &dest, CL_BlendFunc blend_func)
{
	// Currently we only support one blending function.
	cl_assert(blend_func == blend_one_minus_src_alpha);

	lock();
	target->lock();

	char *src_data = (char *) get_data();
	char *dest_data = (char *) target->get_data();

	const CL_PixelFormat &source_format = get_format();
	const CL_PixelFormat &dest_format = target->get_format();

	int src_bytes_per_pixel = (source_format.get_depth()+7)/8;
	int dest_bytes_per_pixel = (dest_format.get_depth()+7)/8;

	src_data += src.top * get_pitch() + src.left * src_bytes_per_pixel;
	dest_data += dest.top * target->get_pitch() + dest.left * dest_bytes_per_pixel;

	// Software c++ template blitter fallback:
	BlitParams params;
	params.src_data = src_data;
	params.dest_data = dest_data;
	params.width = src.get_width();
	params.height = src.get_height();
	params.src_pitch = get_pitch();
	params.dest_pitch = target->get_pitch();
	params.src_red_shift = get_mask_shift(source_format.get_red_mask());
	params.src_green_shift = get_mask_shift(source_format.get_green_mask());
	params.src_blue_shift = get_mask_shift(source_format.get_blue_mask());
	params.src_alpha_shift = get_mask_shift(source_format.get_alpha_mask());
	params.dest_red_shift = get_mask_shift(dest_format.get_red_mask());
	params.dest_green_shift = get_mask_shift(dest_format.get_green_mask());
	params.dest_blue_shift = get_mask_shift(dest_format.get_blue_mask());
	params.dest_alpha_shift = get_mask_shift(dest_format.get_alpha_mask());
	params.src_red_mask = source_format.get_red_mask();
	params.src_green_mask = source_format.get_green_mask();
	params.src_blue_mask = source_format.get_blue_mask();
	params.src_alpha_mask = source_format.get_alpha_mask();
	params.dest_red_mask = dest_format.get_red_mask();
	params.dest_green_mask = dest_format.get_green_mask();
	params.dest_blue_mask = dest_format.get_blue_mask();
	params.dest_alpha_mask = dest_format.get_alpha_mask();

	switch (src_bytes_per_pixel)
	{
	case 1:
		switch (dest_bytes_per_pixel)
		{
		case 1:
			template_blitter<unsigned char, unsigned char>(params);
			break;
		case 2:
			template_blitter<unsigned char, unsigned short>(params);
			break;
		case 4:
			template_blitter<unsigned char, unsigned int>(params);
			break;
		}
		break;
	case 2:
		switch (dest_bytes_per_pixel)
		{
		case 1:
			template_blitter<unsigned short, unsigned char>(params);
			break;
		case 2:
			template_blitter<unsigned short, unsigned short>(params);
			break;
		case 42:
			template_blitter<unsigned short, unsigned int>(params);
			break;
		}
		break;
	case 4:
		switch (dest_bytes_per_pixel)
		{
		case 1:
			template_blitter<unsigned int, unsigned char>(params);
			break;
		case 2:
			template_blitter<unsigned int, unsigned short>(params);
			break;
		case 4:
			template_blitter<unsigned int, unsigned int>(params);
			break;
		}
		break;
	}

	target->unlock();
	unlock();
}
*/
/////////////////////////////////////////////////////////////////////////////
// CL_PixelBuffer implementation:
