/*  $Id: sprite_description.cpp,v 1.18 2003/08/22 12:22:10 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "API/Display/sprite_description.h"
#include "API/Display/pixel_buffer.h"
#include "API/Display/pixel_format.h"
#include "API/Core/System/error.h"
#include "API/Core/Resources/resource_manager.h"
#include "sprite_description_generic.h"
#include "resourcedata_sprite_description.h"

/////////////////////////////////////////////////////////////////////////////
// CL_SpriteDescription construction:

CL_SpriteDescription::CL_SpriteDescription(const std::string &resource_id, CL_ResourceManager *resources)
: impl(0)
{
	CL_Resource &resource = resources->get_resource(resource_id);

	CL_ResourceData_SpriteDescription *data =
		(CL_ResourceData_SpriteDescription *) resource.get_data("sprite_description");

	CL_SpriteDescription desc = data->load_description();
	impl = desc.impl;
	impl->add_ref();
}

CL_SpriteDescription::CL_SpriteDescription()
: impl(new CL_SpriteDescription_Generic)
{
	impl->add_ref();
}

CL_SpriteDescription::CL_SpriteDescription(const CL_SpriteDescription &copy)
: impl(copy.impl)
{
	impl->add_ref();
}

CL_SpriteDescription::~CL_SpriteDescription()
{
	if (impl->release_ref() == 0)
	{
		std::list<CL_PixelBuffer *>::iterator it;
		for(it = impl->delete_providers.begin(); it != impl->delete_providers.end(); ++it)
			delete (*it);
		
		delete impl;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CL_SpriteDescription attributes:

const std::list<CL_SpriteDescription::FramePair> &CL_SpriteDescription::get_frames() const
{
	return impl->frames;
}

CL_Surface::Hint CL_SpriteDescription::get_surface_hint() const
{
	return impl->hint;
}

/////////////////////////////////////////////////////////////////////////////
// CL_SpriteDescription operations:

CL_SpriteDescription &CL_SpriteDescription::operator =(const CL_SpriteDescription &copy)
{
	if (impl->release_ref() == 0)
	{
		std::list<CL_PixelBuffer *>::iterator it;
		for(it = impl->delete_providers.begin(); it != impl->delete_providers.end(); ++it)
			delete (*it);
		
		delete impl;
	}

	impl = copy.impl;
	impl->add_ref();

	return *this;
}

void CL_SpriteDescription::set_surface_hint(CL_Surface::Hint hint)
{
	impl->hint = hint;
}

void CL_SpriteDescription::add_frame(CL_PixelBuffer *provider, bool delete_provider)
{
	provider->lock();
	impl->frames.push_back(FramePair(provider, CL_Rect(0, 0, provider->get_width(), provider->get_height())));
	provider->unlock();
	
	if(delete_provider)
		impl->delete_providers.push_back(provider);
}

void CL_SpriteDescription::add_gridclipped_frames(
	CL_PixelBuffer *provider,
	int xpos, int ypos,
	int width, int height,
	int xarray, int yarray,
	int array_skipframes,
	int xspace, int yspace,
	bool delete_provider)
{
	// TODO: Check if positions are within bounds of provider
	
	int ystart = ypos;
	for(int y = 0; y < yarray; y++)
	{
		int xstart = xpos;
		for(int x = 0; x < xarray; x++)
		{
			if (y == yarray -1 && x >= xarray - array_skipframes)
				break;

			impl->frames.push_back(FramePair(provider, CL_Rect(xstart, ystart, xstart + width, ystart + height)));
			xstart += width + xspace;
		}
		ystart += height + yspace;
	}

	if(delete_provider)
		impl->delete_providers.push_back(provider);
}

void CL_SpriteDescription::add_alphaclipped_frames(CL_PixelBuffer *provider, int xpos, int ypos, float trans_limit, bool delete_provider)
{
	provider->lock();

	CL_PixelFormat format;
	format.set_red_mask(0xff000000);
	format.set_green_mask(0x00ff0000);
	format.set_blue_mask(0x0000ff00);
	format.set_alpha_mask(0x000000ff);
	format.set_depth(32);
	format.enable_colorkey(false);
	
	CL_PixelBuffer alpha_buffer(
		provider->get_width(),
		provider->get_height(),
		provider->get_width()*4,
		format);

	provider->convert(&alpha_buffer);
	provider->unlock();

	int begin=0;
	bool prev_trans=true;

	int alpha_width = alpha_buffer.get_width();
	int alpha_height = alpha_buffer.get_height();
	bool found_opaque = false;
	bool found_trans = false;

	int *opaque_row = new int[alpha_width];
	memset(opaque_row, 0, alpha_width*sizeof(int));

	int cut_top = 0;
	int cut_bottom = alpha_height;
		
	alpha_buffer.lock();
	char *data = (char *) alpha_buffer.get_data();
		
	for (int y=0; y < alpha_height; y++)
	{
		bool opaque_line = false;
		unsigned int *line = (unsigned int *) (data + alpha_buffer.get_pitch()*y);
		for (int x=0; x < alpha_width; x++)
		{
			if ((line[x] & 255) > trans_limit*255)
			{
				opaque_row[x] = 1;
				opaque_line = true;
				found_opaque = true;
			}
		}
			
		if (opaque_line == false) // cut something of top or bottom
		{
			if (found_opaque)
			{
				cut_bottom--;
				found_trans = true;
			}
			else
				cut_top ++;
		}
		else if (found_trans)
		{
			found_trans = false;
			cut_bottom = alpha_height;
		}
	}
	
	alpha_buffer.unlock();

	if (cut_top >= cut_bottom)
		throw CL_Error("Font image contained only alpha!");

	for(int x=0; x < alpha_width; x++)
	{
		if(opaque_row[x] && prev_trans)
		{
			begin = x;
			prev_trans = false;
		}
		else if (!opaque_row[x] && !prev_trans)
		{
			impl->frames.push_back(
				FramePair(provider, CL_Rect(begin, cut_top, x+1, cut_bottom)));

			prev_trans = true;
		}
	}
		
	if (!prev_trans)
	{
		impl->frames.push_back(
			FramePair(provider, CL_Rect(begin, cut_top, alpha_width, cut_bottom)));
	}

	delete[] opaque_row;

	if(delete_provider)
		impl->delete_providers.push_back(provider);
}

void CL_SpriteDescription::add_paletteclipped_frames(CL_PixelBuffer *provider, int xpos, int ypos, bool delete_provider)
{
	provider->lock();

	if (provider->get_format().get_depth() != 8)
	{
		provider->unlock();
		throw CL_Error("Paletteclipped frames only works on PAL8 (256 colours) images.");
	}
		
	unsigned char *surface_data = (unsigned char *) provider->get_data();

	// Find height
	int height;
	for (height = ypos; height < provider->get_height(); height++)
	{
		if (surface_data[height * provider->get_pitch() + xpos] == 255)
		{
			height -= ypos;
			break;
		}
	}
	
	int org_x = xpos;
	bool next_line = false;
	bool found;
	
	do
	{
		// Find width
		int width;
		int provider_width = provider->get_width();
		found = false;
		for (width = xpos; width < provider_width; width++)
		{
			if (surface_data[ypos * provider->get_pitch() + width] == 254)
			{
				found = true;
				width -= xpos;
				break;
			}
			else if (surface_data[ypos * provider->get_pitch() + width] == 253)
			{
				found = true;
				width -= xpos;
				next_line = true;
				break;
			}
		}
		
		// Cut frame if found
		if(found)
		{
			impl->frames.push_back(FramePair(provider, CL_Rect(xpos, ypos, xpos + width, ypos + height)));
			
			xpos += width + 1;

			if (next_line)
			{
				xpos = org_x;
				ypos += height + 1;
				next_line = false;
			}
		}
	} while(found);

	provider->unlock();

	if(delete_provider)
		impl->delete_providers.push_back(provider);
}

/////////////////////////////////////////////////////////////////////////////
// CL_SpriteDescription implementation:
