/*  $Id: gl_state_checker.cpp,v 1.30 2003/10/12 21:47:57 sphair Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "GL/gl_state.h"
#include "GL/gl_state_checker.h"
#include "API/Display/blend_func.h"
#include "API/Display/color.h"
#include "API/Core/Math/rect.h"
#include "API/Core/Math/point.h"

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateChecker defines, statics

// Number of quads that can be drawn at once.
// If more are drawn, a flush() will take place.
#define MAX_QUAD_COUNT 256

// Need to distinguish gradients from plain quads.
// This can be whatever except GL_QUADS, GL_LINES or GL_POINTS.
#define CL_GRADIENT 0x1337

static int primitive_count = 0;  // keep track of how many primitives to draw
static GLfloat vertex_array[MAX_QUAD_COUNT*3*4];
static GLint texcoord_array[MAX_QUAD_COUNT*2*4];
static GLfloat color_array[MAX_QUAD_COUNT*4*4];

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateChecker construction:

CL_GLStateChecker::CL_GLStateChecker()
{
	internal_state.blendfunc[0] = blend_src_alpha;
	internal_state.blendfunc[1] = blend_one_minus_src_alpha;
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	internal_state.color[0] = 255;
	internal_state.color[1] = 255;
	internal_state.color[2] = 255;
	internal_state.color[3] = 255;
	glColor4fv(internal_state.color);

	// -2 since we want to force the first render to set enable/disable_2d
	internal_state.texture_handle = -2;
	
	glDisable(GL_TEXTURE_2D);

	internal_state.rotate_angle = 0.0f;
	
	primitive = GL_DONT_CARE;
}

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateChecker operations:

void CL_GLStateChecker::add_clear(
	const CL_Color &color)
{
	flush();
	
	glClearColor(color.get_red() / 255.0f, color.get_green() / 255.0f, color.get_blue() / 255.0f, color.get_alpha() / 255.0f);
	glClear(GL_COLOR_BUFFER_BIT);
}

void CL_GLStateChecker::add_pixel(
	const CL_GLState &state,
	const CL_Pointf &position)
{
	set_state(state, GL_POINTS);

	glVertex2f(position.x + 0.375f, position.y + 0.375f);
}

void CL_GLStateChecker::add_line(
	const CL_GLState &state,
	const CL_Rectf &position)
{
	set_state(state, GL_LINES);

	glVertex2f(position.left + 0.375f, position.top + 0.375f);
	glVertex2f(position.right + 0.375f, position.bottom + 0.375f);
}

void CL_GLStateChecker::add_gradient(
	const CL_GLState &state,
	const CL_Rectf &position,
	const CL_Gradient &gradient)
{
	set_state(state, CL_GRADIENT);

	add_quad_vertices( state, position.left, position.top, position.right, position.bottom );

	int a = primitive_count*16;

	// Bottom left vertex
	color_array[a++] = gradient.bottom_left.get_red() / 255.0f;
	color_array[a++] = gradient.bottom_left.get_green() / 255.0f;
	color_array[a++] = gradient.bottom_left.get_blue() / 255.0f;
	color_array[a++] = gradient.bottom_left.get_alpha() / 255.0f;

	// Bottom right vertex
	color_array[a++]  = gradient.bottom_right.get_red() / 255.0f;
	color_array[a++]  = gradient.bottom_right.get_green() / 255.0f;
	color_array[a++] =	gradient.bottom_right.get_blue() / 255.0f;
	color_array[a++] =	gradient.bottom_right.get_alpha() / 255.0f;

	// Top right vertex
	color_array[a++] = gradient.top_right.get_red() / 255.0f;
	color_array[a++] = gradient.top_right.get_green() / 255.0f;
	color_array[a++] = gradient.top_right.get_blue() / 255.0f;
	color_array[a++] = gradient.top_right.get_alpha() / 255.0f;

	// Top left vertex
	color_array[a++] = gradient.top_left.get_red() / 255.0f;
	color_array[a++] = gradient.top_left.get_green() / 255.0f;
	color_array[a++] = gradient.top_left.get_blue() / 255.0f;
	color_array[a++] = gradient.top_left.get_alpha() / 255.0f;

	primitive_count++;

	if( primitive_count == MAX_QUAD_COUNT )
		flush();
}

void CL_GLStateChecker::add_quad_vertices( 
	const CL_GLState &state,
	float px1,
	float py1,
	float px2,
	float py2 )
{
	int b = primitive_count * 8; // 8 values per quad
	
	if (state.rotate_angle == 0.0f)
	{
		vertex_array[b+0] = px1;
		vertex_array[b+1] = py2;
		vertex_array[b+2] = px2;
		vertex_array[b+3] = py2;
		vertex_array[b+4] = px2;
		vertex_array[b+5] = py1;
		vertex_array[b+6] = px1;
		vertex_array[b+7] = py1;
	}
	else
	{
		vertex_array[b+0] = calc_rotate_x(px1, py2);
		vertex_array[b+1] = calc_rotate_y(px1, py2);
		vertex_array[b+2] = calc_rotate_x(px2, py2);
		vertex_array[b+3] = calc_rotate_y(px2, py2);
		vertex_array[b+4] = calc_rotate_x(px2, py1);
		vertex_array[b+5] = calc_rotate_y(px2, py1);
		vertex_array[b+6] = calc_rotate_x(px1, py1);
		vertex_array[b+7] = calc_rotate_y(px1, py1);
	}
}


void CL_GLStateChecker::add_quad(
	const CL_GLState &state,
	float px1,
	float py1,
	float px2,
	float py2,
	int tx1,
	int ty1,
	int tx2,
	int ty2)
{
	set_state(state, GL_QUADS);

	if( internal_state.texture_handle != -1 )
	{
		int a = primitive_count * 8; // offset in array. 8 values per quad 
		
		texcoord_array[a+0] = tx1;
		texcoord_array[a+1] = ty2;
		texcoord_array[a+2] = tx2;
		texcoord_array[a+3] = ty2;
		texcoord_array[a+4] = tx2;
		texcoord_array[a+5] = ty1;
		texcoord_array[a+6] = tx1;
		texcoord_array[a+7] = ty1;
	}

	add_quad_vertices( state, px1, py1, px2, py2 );
	
	primitive_count++;
		
	if( primitive_count == MAX_QUAD_COUNT )
		flush();
}

void CL_GLStateChecker::flush()
{
	if (primitive == GL_LINES || primitive == GL_POINTS )
	{
		glEnd();
		primitive = GL_DONT_CARE; // carefull... GL_POINTS is defined as 0x0000
	}

	if( primitive_count != 0 )
	{
		// pointer must be set every frame, otherwise kabooom!
		glVertexPointer( 2, GL_FLOAT, 0, vertex_array );
		glTexCoordPointer( 2, GL_INT, 0, texcoord_array );

		// enable color array if gradient
		if( primitive == CL_GRADIENT )
		{
			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer( 4, GL_FLOAT, 0, color_array );
		}

		if( primitive == GL_QUADS || primitive == CL_GRADIENT )
		{
			glDrawArrays(GL_QUADS, 0, 4*primitive_count);
		}
		
		if( primitive == CL_GRADIENT ) // not gradient any more. disable color array.
		{
			glDisableClientState(GL_COLOR_ARRAY);
			glColor4fv(internal_state.color);
		}

		primitive_count = 0;
	}
}

#ifdef WIN32

void CL_GLStateChecker::set_context(HDC hdc, HGLRC context)
{
	if (cur_drawable != hdc && cur_context != context)
	{
		flush();
		wglMakeCurrent(hdc, context);
		cur_drawable = hdc;
		cur_context = context;
	}
}

#else

void CL_GLStateChecker::set_context(Display *disp, GLXDrawable drawable, GLXContext context)
{
	if (cur_drawable != drawable && cur_context != context)
	{
		flush();
		glXMakeCurrent(disp, drawable, context);
		cur_drawable = drawable;
		cur_context = context;
	}
}

#endif

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateTree implementation:

void CL_GLStateChecker::set_state(const CL_GLState &state, GLenum new_primitive)
{
	if (internal_state.texture_handle != state.texture_handle)
	{
		internal_state.texture_handle = state.texture_handle;
		
		flush();

		if(state.texture_handle == -1)
		{
			glDisable(GL_TEXTURE_2D);
		}
		else
		{
			// Bind texture and convert internal texture coordinates to be
			// in pixels.
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, (GLuint)state.texture_handle);
			GLint texture_width = 1, texture_height = 1;
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &texture_width);
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &texture_height);
			glMatrixMode(GL_TEXTURE);
			GLdouble scale_matrix[16];
			memset(scale_matrix, 0, sizeof(GLdouble)*16);
			scale_matrix[0] = 1.0/texture_width;
			scale_matrix[5] = 1.0/texture_height;
			scale_matrix[10] = 1.0;
			scale_matrix[15] = 1.0;
			glLoadMatrixd(scale_matrix);
			glMatrixMode(GL_MODELVIEW);
		}
	}
	
	if (internal_state.blendfunc[0] != state.blendfunc[0] || 
		internal_state.blendfunc[1] != state.blendfunc[1])
	{
		flush();

		internal_state.blendfunc[0] = state.blendfunc[0];
		internal_state.blendfunc[1] = state.blendfunc[1];
		
		GLenum blendfunc[2];
		for(int i = 0; i < 2; ++i)
		{			
			switch (state.blendfunc[i])
			{
			case blend_zero:
				blendfunc[i] = GL_ZERO;
				break;
			case blend_one:
				blendfunc[i] = GL_ONE;
				break;
			case blend_dest_color:
				blendfunc[i] = GL_DST_COLOR;
				break;
			case blend_src_color:
				blendfunc[i] = GL_SRC_COLOR;
				break;
			case blend_one_minus_dest_color:
				blendfunc[i] = GL_ONE_MINUS_DST_COLOR;
				break;
			case blend_one_minus_src_color:
				blendfunc[i] = GL_ONE_MINUS_SRC_COLOR;
				break;
			case blend_src_alpha:
				blendfunc[i] = GL_SRC_ALPHA;
				break;
			case blend_one_minus_src_alpha:
				blendfunc[i] = GL_ONE_MINUS_SRC_ALPHA;
				break;
			case blend_dst_alpha:
				blendfunc[i] = GL_DST_ALPHA;
				break;
			case blend_one_minus_dest_alpha:
				blendfunc[i] = GL_ONE_MINUS_DST_ALPHA;
				break;
			case blend_src_alpha_saturate:
				blendfunc[i] = GL_SRC_ALPHA_SATURATE;
				break;
			case blend_constant_color:
				break;
			case blend_constant_alpha:
				break;
			case blend_one_minus_constant_color:
				break;
			case blend_one_minus_constant_alpha:
				break;
			}
		}
		glBlendFunc(blendfunc[0], blendfunc[1]);
	}

	if (internal_state.color[0] != state.color[0] ||
		internal_state.color[1] != state.color[1] ||
		internal_state.color[2] != state.color[2] ||
		internal_state.color[3] != state.color[3])
	{
		flush();

		internal_state.color[0] = state.color[0];
		internal_state.color[1] = state.color[1];
		internal_state.color[2] = state.color[2];
		internal_state.color[3] = state.color[3];

		glColor4fv(state.color);
	}

	if ( primitive != new_primitive  )
	{
		flush();

		if ( new_primitive == GL_LINES || new_primitive == GL_POINTS )
			glBegin(new_primitive);

		primitive = new_primitive;
	}

	internal_state.rotation_hotspot = state.rotation_hotspot;

	if (internal_state.rotate_angle != state.rotate_angle)
	{
		internal_state.rotate_angle = state.rotate_angle;

		double angle_rad = 3.1415926f*state.rotate_angle/180;
		rotate_x.x = cos(angle_rad);
		rotate_x.y = sin(angle_rad);
		rotate_y.x = cos(3.1415926f/2+angle_rad);
		rotate_y.y = sin(3.1415926f/2+angle_rad);
	}
}

CL_Pointf CL_GLStateChecker::rotate_x(1.0f, 0.0f);

CL_Pointf CL_GLStateChecker::rotate_y(0.0f, 1.0f);

CL_GLStateChecker gl_state_tree;

CL_GLState CL_GLStateChecker::internal_state;

GLenum CL_GLStateChecker::primitive;

#ifdef WIN32
	HDC CL_GLStateChecker::cur_drawable = 0;

	HGLRC CL_GLStateChecker::cur_context = 0;
#else
	GLXDrawable CL_GLStateChecker::cur_drawable = 0;

	GLXContext CL_GLStateChecker::cur_context = 0;
#endif
