/*  $Id: graphic_context_opengl.h,v 1.17 2003/10/14 15:52:12 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef header_graphic_context_opengl
#define header_graphic_context_opengl

#if _MSC_VER > 1000
#pragma once
#endif

#include "Display/graphic_context_generic.h"
#include "API/signals.h"

class CL_DisplayWindow_Generic;
class CL_Surface_Target_OpenGL;

//: OpenGL graphic context
class CL_GraphicContext_OpenGL : public CL_GraphicContext_Generic
{
//! Construction:
public:
	CL_GraphicContext_OpenGL(CL_DisplayWindow_Generic *window);

	CL_GraphicContext_OpenGL(CL_Surface_Target_OpenGL *surface_target);

	virtual ~CL_GraphicContext_OpenGL();

//! Attributes:
public:
	//: Returns the current width of the context.
	virtual int get_width() const;

	//: Returns the current height of the context.
	virtual int get_height() const;

	//: Returns a pixel value at specified coordinates.
	virtual CL_Color get_pixel(int x, int y) const;

//! Operations:
public:
	//: Called prior to every drawal command.
	virtual void select_gc();

	//: Draw a pixel at (x, y) using the specified color.
	virtual void draw_pixel(int x, int y, const CL_Color &color);

	//: Draw a line from (x1, y1) to (x2, y2) using the specified color.
	virtual void draw_line(int x1, int y1, int x2, int y2, const CL_Color &color);

	//: Draw a rectangle using the specified color.
	virtual void draw_rect(const CL_Rect &rect, const CL_Color &color);

	//: Draw a gradient rectangle using the specified gradient.
	virtual void fill_rect(const CL_Rect &rect, const CL_Gradient &gradient);

	//: Draw a filled rectangle using the specified color.
	virtual void fill_rect(const CL_Rect &rect, const CL_Color &color);

	//: Clears the whole context using the specified color.
	virtual void clear(const CL_Color &color);

	//: Set the current clipping rectangle.
	virtual void set_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to stack; then set clipping rectangle to the
	//: union of the current rectangle and the passed rectangle.
	virtual void push_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to the stack.
	virtual void push_cliprect();

	//: Pop current clipping rectangle from the stack.
	virtual void pop_cliprect();

	//: Sets the effective translate offset to (x,y).
	virtual void set_translate_offset(int x, int y);
	
	//: Push translation offset onto translation stack.
	//- <p>This offset will
	//- affect any subsequent display operations on the current displaycard, by
	//- translating the position of the display operation with the offset.
	//- The offset will be offset by any previous offsets pushed onto the stack,
	//- eg. it inherits the previous offset.</p>
	virtual void push_translate_offset(int delta_x, int delta_y);
	
	//: Sets the translation offset as a new absolute translation offset.
	//- <p>The new offset will disregard any previous offset's, but will not
	//- empty the translation stack. The new translation offset will affect
	//- any subsequent display operations on the current displaycard, by
	//- translating the position of the display operation with the offset.</p>
	virtual void push_translate_offset();
	
	//: Pops the last pushed translation offset from the translation offset
	//: stack.
	//- <p>If the stack is empty, nothing will happen, and if the last
	//- translation offset is popped, the translation offset will be set to 0,0.</p>
	virtual void pop_translate_offset();

	//: Flushes current rendering batch.
	//- <p>With the OpenGL target, this causes the graphic context to end its current
	//- active glBegin()/glEnd() pair, making it possible to call other OpenGL calls.</p>
	virtual void flush();

	//: Called when 2D operations are begun.
	virtual void setup_2d();

	//: Called when 2D operations are ended.
	virtual void setup_3d();

//! Implementation:
private:
	//: Updates glScissor with current clip rect.
	void update_scissor();

	//: Called when window is resized.
	void on_resize(int width, int height);

	//: True when currently in 2d mode.
	bool in_2d;

	//: OpenGL window.
	//- Can be null if attached to a surface instead.
	CL_DisplayWindow_Generic *window;

	//: OpenGL surface.
	//- Can be null if attached to a window instead.
	CL_Surface_Target_OpenGL *surface_target;

	//: Slots.
	CL_SlotContainer slots;
};

#endif
