/*  $Id: surface_target_opengl.cpp,v 1.5 2003/09/30 10:33:47 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "surface_target_opengl.h"
#include "gl_state_checker.h"
#include "API/Display/pixel_buffer.h"
#include "API/Display/graphic_context.h"
#include "API/Display/pixel_format.h"
#include <Hermes/H_Conv.h>

#include <GL/glu.h>

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL construction:

CL_Surface_Target_OpenGL::CL_Surface_Target_OpenGL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Surface::Hint hint)
{
	CL_GLStateChecker::flush();

	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	provider->lock();

	screen_size.width = (float)provider->get_width();
	screen_size.height = (float)provider->get_height();
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < screen_size.width) texture_size.width *= 2;
	while (texture_size.height < screen_size.height) texture_size.height *= 2;

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		static_cast<int>(texture_size.width),
		static_cast<int>(texture_size.height),
		static_cast<int>(texture_size.width) * 4,
		abgr_format);

	provider->convert(
		buffer->get_data(), abgr_format,
		buffer->get_pitch(),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)));

	provider->unlock();
	if (delete_provider) delete provider;

	buffer->lock();
	
	// Detect what level of alpha we require if application didnt tell us:
	if (hint == CL_Surface::hint_automatic)
	{
		bool found_colorkey = false;
		bool found_alpha = false;
	
		unsigned char *data = (unsigned char *) buffer->get_data();
		for (int y=0; y<screen_size.height; y++)
		{
			for (int x=0; x<screen_size.width; x++)
			{
				unsigned char alpha = data[x*4+3];
				if (alpha == 0) found_colorkey = true;
				else if (alpha != 255) found_alpha = true;
			}
			data += buffer->get_pitch();
		}
		
		if (found_alpha) hint = CL_Surface::hint_alpha;
		else if (found_colorkey) hint = CL_Surface::hint_colorkey;
		else hint = CL_Surface::hint_opaque;
	}
	
	// Figure out what is the optimal supported format for this card:
	GLint desired_format = GL_RGBA8;
	if (hint == CL_Surface::hint_opaque) desired_format = GL_RGB8;
	glTexImage2D(
		GL_PROXY_TEXTURE_2D,      // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		GL_RGBA,                  // format
		GL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	GLint obtained_format = desired_format;
	glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_INTERNAL_FORMAT, &obtained_format);
	switch (hint)
	{
	case CL_Surface::hint_alpha:
		break;
	
	case CL_Surface::hint_colorkey:
		// If we got 16 bit, try to get 5 bits for colors if possible.
		if (obtained_format == GL_RGBA4) desired_format = GL_RGB5_A1;
		break;
	
	case CL_Surface::hint_opaque:
		break;

	case CL_Surface::hint_automatic:
		break;
	}

	// Upload to OpenGL:
	glGenTextures(1, &handle);
	glBindTexture(GL_TEXTURE_2D, handle);

	glTexImage2D(
		GL_TEXTURE_2D,            // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		GL_RGBA,                  // format
		GL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

	buffer->unlock();
	delete buffer;
}

CL_Surface_Target_OpenGL::~CL_Surface_Target_OpenGL()
{
	glDeleteTextures(1, &handle);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL attributes:

int CL_Surface_Target_OpenGL::get_width() const
{
	return (int)screen_size.width;
}
	
int CL_Surface_Target_OpenGL::get_height() const
{
	return (int)screen_size.height;
}

CL_PixelBuffer CL_Surface_Target_OpenGL::get_pixeldata()
{
	return CL_PixelBuffer();
}
	
CL_GraphicContext *CL_Surface_Target_OpenGL::get_gc()
{
	return 0;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL operations:

void CL_Surface_Target_OpenGL::set_pixeldata(
	const CL_Point &pos,
	const CL_Rect &src_rect,
	CL_PixelBuffer &pixeldata)
{
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	int x, int y,
	CL_GraphicContext *gc)
{
	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		screen_size);

	init_attributes(attributes, screen_size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		float x_pos = x - translation.x;
		float y_pos = y - translation.y;
		state.rotation_hotspot.x += x_pos;
		state.rotation_hotspot.y += y_pos;

		CL_GLStateChecker::add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + static_cast<int>(screen_size.width),
			y_pos + static_cast<int>(screen_size.height),
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
	else
	{
		float x_pos = x - translation.x * attributes->scale_x;
		float y_pos = y - translation.y * attributes->scale_y;
		state.rotation_hotspot.x = x_pos + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = y_pos + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + screen_size.width * attributes->scale_x,
			y_pos + screen_size.height * attributes->scale_y,
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	CL_Sizef size((float) dest.get_width(), (float) dest.get_height());
	
	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

//	CL_GLState state = init_attributes(attributes, size);
	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	CL_Sizef size((float)dest.get_width(), (float)dest.get_height());

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
}

void CL_Surface_Target_OpenGL::select_context()
{
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL implementation:

/*
	glBindTexture(GL_TEXTURE_2D, handle);

	// Store matrix
	glPushMatrix();

	// Calculate translation hotspot
	CL_Point translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		CL_Size(width, height));

	// Translate to translation hotspot
	glTranslatef(-translation.x + x, -translation.y + y, 0.0f);

	// Scale surface
	if(attributes->scale_x != 1.0f || attributes->scale_y != 1.0f)
	{
		// Translate back to origin
		glTranslatef(translation.x, translation.y, 0.0f);

		// Scale surface
		glScalef(attributes->scale_x, attributes->scale_y, 1.0f);

		// Translate for scaling
		glTranslatef(-translation.x, -translation.y, 0.0f);
	}

	// Rotate surface
	if(attributes->angle)
	{
		// Calculate rotation hotspot
		CL_Point rotation = calc_hotspot(
			attributes->rotation_origin,
			attributes->rotation_hotspot,
			CL_Size(width, height));

		// Translate back to origin
		glTranslatef(rotation.x, rotation.y, 0.0f);

		// Rotate surface
		glRotatef(attributes->angle, 0, 0, 1);

		// Translate to rotation hotspot
		glTranslatef(-rotation.x, -rotation.y, 0.0f);
	}

	// TODO: Set ALL blend func
	if(attributes->blend_src == blend_src_alpha && attributes->blend_dest == blend_one)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
*/
