/*  $Id: surface_sdl.cpp,v 1.10 2003/10/16 11:00:28 sphair Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "surface_sdl.h"
#include "API/Display/pixel_buffer.h"
#include "API/Display/palette.h"
#include "API/Display/graphic_context.h"
#include "API/Display/pixel_format.h"
#include "API/Core/System/error.h"
#include "API/Core/System/log.h"
#include "API/core.h"
#include <Hermes/H_Conv.h>


/////////////////////////////////////////////////////////////////////////////
// CL_Surface_SDL construction:

CL_Surface_SDL::CL_Surface_SDL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Surface::Hint hint)
{
	Uint32 mode = SDL_SWSURFACE;
	surface = 0;
	SDL_Surface *tmp = 0;

	provider->lock();

	tmp = SDL_CreateRGBSurfaceFrom(provider->get_data(),
				       provider->get_width(),
				       provider->get_height(),
				       provider->get_format().get_depth(),
				       provider->get_pitch(),
				       provider->get_format().get_red_mask(),
				       provider->get_format().get_green_mask(),
				       provider->get_format().get_blue_mask(),
				       provider->get_format().get_alpha_mask());

	if(tmp == 0)
		throw CL_Error(SDL_GetError());

	if(tmp->format->BitsPerPixel == 8){
		const CL_Color *palette_colors = provider->get_palette().colors;
		tmp->format->palette->ncolors = 256;
		int i;
		for(i=0;i<256;i++){
			tmp->format->palette->colors[i].r = palette_colors[i].get_red();
			tmp->format->palette->colors[i].g = palette_colors[i].get_green();
			tmp->format->palette->colors[i].b = palette_colors[i].get_blue();
			tmp->format->palette->colors[i].unused = 0;
		}

		surface = SDL_CreateRGBSurface(SDL_SWSURFACE,
					       provider->get_width(),
					       provider->get_height(),
					       provider->get_format().get_depth(),
					       provider->get_format().get_red_mask(),
					       provider->get_format().get_green_mask(),
					       provider->get_format().get_blue_mask(),
					       provider->get_format().get_alpha_mask());
		
		surface->format->palette->ncolors = 256;
		for(i=0;i<256;i++){
			surface->format->palette->colors[i].r = palette_colors[i].get_red();
			surface->format->palette->colors[i].g = palette_colors[i].get_green();
			surface->format->palette->colors[i].b = palette_colors[i].get_blue();
			surface->format->palette->colors[i].unused = 0;
		}
		
		SDL_BlitSurface(tmp,0,surface,0);
		SDL_FreeSurface(tmp);
		/* Uncomment this if you want to convert 8bits surfaces to screen format display
		   The conversion itself is slow though once it is done it increases the blitting performance
		tmp = SDL_DisplayFormat(surface);
		SDL_FreeSurface(surface);
		surface = tmp;
		*/
	}else{
		if(tmp->format->BitsPerPixel == 32){
			// Detect what level of alpha we require if application didnt tell us:
			if (hint == CL_Surface::hint_automatic)
			{
				bool found_colorkey = false;
				bool found_alpha = false;
				
				unsigned char *data = (unsigned char *) provider->get_data();
				for (int y=0; y<provider->get_height(); y++)
				{
					for (int x=0; x<provider->get_width(); x++)
					{
						unsigned char alpha = data[x*4+3];
						if (alpha == 0) found_colorkey = true;
						else if (alpha != 255) found_alpha = true;
					}
					data += provider->get_pitch();
				}
				
				if (found_alpha) hint = CL_Surface::hint_alpha;
				else if (found_colorkey) hint = CL_Surface::hint_colorkey;
				else hint = CL_Surface::hint_opaque;
			}
			switch (hint){
			case CL_Surface::hint_alpha:
				mode = SDL_SRCALPHA|SDL_HWSURFACE;
				break;
			case CL_Surface::hint_colorkey:
				mode = SDL_SRCCOLORKEY|SDL_HWSURFACE;
				break;
			case CL_Surface::hint_opaque:
				mode = SDL_SRCALPHA|SDL_HWSURFACE;
				break;
			case CL_Surface::hint_automatic:
				break;
			}
		}

		surface = SDL_ConvertSurface(tmp,tmp->format,mode);
		SDL_FreeSurface(tmp);
	}

	if(provider->get_format().has_colorkey())
		SDL_SetColorKey(surface,SDL_RLEACCEL|SDL_SRCCOLORKEY,provider->get_format().get_colorkey());

        if (delete_provider) 
		delete provider;

}

CL_Surface_SDL::~CL_Surface_SDL()
{
	if(surface != 0)
		SDL_FreeSurface(surface);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_SDL attributes:

int CL_Surface_SDL::get_width() const
{
	return surface->w;
}
	
int CL_Surface_SDL::get_height() const
{
	return surface->h;
}

CL_PixelBuffer CL_Surface_SDL::get_pixeldata()
{
	return CL_PixelBuffer();
}
	
CL_GraphicContext *CL_Surface_SDL::get_gc()
{
	return 0;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_SDL operations:

void CL_Surface_SDL::set_pixeldata(
	const CL_Point &pos,
	const CL_Rect &src_rect,
	CL_PixelBuffer &pixeldata)
{
}

//: Draw surface on screen.
void CL_Surface_SDL::draw(
	CL_Surface_Generic *attributes,
	int x, int y,
	CL_GraphicContext *gc)
{
	// todo: update dst rect to take gc->get_cliprect() into account.

	x += gc->get_translate_x();
	y += gc->get_translate_y();
	SDL_Rect dst = {x,y,surface->w,surface->h};
	if(surface->format->BitsPerPixel > 8)
	{
		Uint8 a = (Uint8)(attributes->alpha * 255);
		SDL_SetAlpha(surface,SDL_RLEACCEL|SDL_SRCALPHA,a);
	}
	SDL_BlitSurface(surface,0,SDL_GetVideoSurface(),&dst);
}

//: Draw surface on screen.
void CL_Surface_SDL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	// todo: update dst rect to take gc->get_cliprect() into account.
	
	int x,y;
	x = dest.left + gc->get_translate_x();
	y = dest.top + gc->get_translate_y();
	SDL_Rect dst = {x,y,dest.get_width(),dest.get_height()};
	if(surface->format->BitsPerPixel > 8)
	{
		Uint8 a = (Uint8)(attributes->alpha * 255);
		SDL_SetAlpha(surface,SDL_RLEACCEL|SDL_SRCALPHA,a);
	}
	SDL_BlitSurface(surface,0,SDL_GetVideoSurface(),&dst);
}

//: Draw surface on screen.
void CL_Surface_SDL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	// todo: update dst rect to take gc->get_cliprect() into account.

	int x,y;
	x = dest.left + gc->get_translate_x();
	y = dest.top + gc->get_translate_y();
	SDL_Rect source = {src.left,src.top,src.get_width(),src.get_height()};
	SDL_Rect dst = {x,y,dest.get_width(),dest.get_height()};
	if(surface->format->BitsPerPixel > 8)
	{
		Uint8 a = (Uint8)(attributes->alpha * 255);
		SDL_SetAlpha(surface,SDL_RLEACCEL|SDL_SRCALPHA,a);
	}
	SDL_BlitSurface(surface,&source,SDL_GetVideoSurface(),&dst);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_SDL implementation:
